# ---------------- Introduction ---------------- #

# This is a Python (v3.8.6) script written by Filip Arnaut, Laboratory for Astrophysics and Physics of Ionosphere,
# Institute of Physics Belgrade, University of Belgrade, Republic of Serbia.

# This script utilizes the PyCaret library for machine learning classification of VLF amplitude ionospheric signals
# for the classification of erroneous data, instrumental malfunction, solar flare influence and overall noisy data
# from normal VLF amplitude signals.

# More information can be found in:

# Arnaut F, Kolarski A, Srećković VA. Random Forest Classification and Ionospheric Response to Solar Flares:
# Analysis and Validation. Universe. 2023; 9(10):436. https://doi.org/10.3390/universe9100436

# ---------------- Import libraries ---------------- #

import pandas as pd
import os
from pycaret.classification import *
from sklearn.metrics import confusion_matrix
import matplotlib.pyplot as plt
from tabulate import tabulate

# ---------------- Parameter settings ---------------- #

# Load the train dataset as a csv file
train_dataset= pd.read_csv('train_dataset.csv')

# Load the test dataset as a csv file
test_dataset = pd.read_csv('test_dataset.csv')

# Define the range for data visualization
range_1, range_2= 700, 1300

# ---------------- ML Classification function ---------------- #

def classification():
    # Load train and test datasets
    train_data = train_dataset
    test_data = test_dataset

    # Features and target
    # Features can be removed by just deleting a given feature from the list
    features = ['timeLOC_float',
                'roll_stdev_180_ampl_orig',
                'roll_stdev_20_X_ray',
                'roll_mean_180_ampl_orig',
                'roll_median_180_ampl_orig',
                'roll_stdev_180_X_ray',
                'roll_median_180_X_ray',
                'roll_mean_180_X_ray',
                'roll_mean_20_ampl_orig',
                'roll_median_20_ampl_orig',
                'roll_median_5_ampl_orig',
                'roll_stdev_20_ampl_orig',
                'roll_mean_5_ampl_orig',
                'lagged_3_ampl_orig',
                'X_ray',
                'lagged_1_ampl_orig',
                'ampl_orig',
                'roll_mean_5_X_ray',
                'lagged_1_X_ray',
                'lagged_4_ampl_orig']

    # Single target for classification
    target = 'target'  

    # Initialize PyCaret setup
    clf1 = setup(data=train_data[features + [target]], target=target)

    # Compare all the available models and find the best one
    best = compare_models()

    # Create and tune the model according to the best model from the train dataset
    model = create_model(best)
    tuned_model = tune_model(model)

    # Predict on the test dataset
    predictions = predict_model(tuned_model, data=test_data[features])

    # Add the actual target variable to the predictions DataFrame
    predictions['target'] = test_data[target]
    predictions['transmitter'] = test_data['transmitter']
    predictions['rec'] = test_data['rec']
    predictions['date'] = test_data['date']
    predictions['timeUT'] = test_data['timeUT']

    # Save the DataFrame to a CSV file
    predictions.to_csv('output_predictions.csv', index=False)

    return predictions

# ---------------- Post-processing functions ---------------- #

def decode_trans_rec_data(input_file):
    # Load the CSV file into a DataFrame
    df = pd.read_csv(input_file, low_memory=False)

    # Define mappings for 'transmitter' column
    transmitter_map = {1: 'NAA', 2: 'NAU', 3: 'NLK', 4: 'NML', 5: 'NPM'}
    df['transmitter'] = df['transmitter'].map(transmitter_map)

    # Define mappings for 'rec' column
    rec_map = {1: 'OklahomaEast', 2: 'OklahomaSouth', 3: 'Sheridan', 4: 'Walsenburg'}
    df['rec'] = df['rec'].map(rec_map)

    # Save the transformed DataFrame to a new CSV file
    df.to_csv('transformed_data.csv', index=False)

def unmerge_trans_rec_pairs(csv_file_path="transformed_data.csv", output_folder="graphing_data"):
    # Load the CSV file
    df = pd.read_csv(csv_file_path)
    
    # Get the absolute path of the current script
    script_path = os.path.abspath(__file__)
    # Get the directory containing the script
    script_directory = os.path.dirname(script_path)
    # Combine the script directory with the output folder name
    output_folder_path = os.path.join(script_directory, output_folder)
    
    # Create the output folder if it doesn't exist
    os.makedirs(output_folder_path, exist_ok=True)
    
    # Group the DataFrame by transmitter and receiver
    grouped_df = df.groupby(['transmitter', 'rec'])
    
    # Iterate over each group
    for (transmitter, receiver), group_df in grouped_df:
        # Generate the file name based on transmitter and receiver
        file_name = f"{transmitter}_{receiver}.csv"
        
        # Save the group DataFrame to a CSV file in the output folder
        group_df.to_csv(os.path.join(output_folder_path, file_name), index=False)

def calculate_evaluation_metrics(csv_folder='graphing_data', output_file='evaluation_metrics.txt'):
    # Get the absolute path of the current script
    script_path = os.path.abspath(__file__)
    # Get the directory containing the script
    script_directory = os.path.dirname(script_path)
    # Combine the script directory with the output file name
    output_file_path = os.path.join(script_directory, output_file)

    with open(output_file_path, 'w') as txt_file:
        # Get the list of CSV files in the folder
        file_list = [f for f in os.listdir(csv_folder) if f.endswith('.csv')]

        # Iterate over each file
        for file_name in file_list:
            # Load the CSV file
            file_path = os.path.join(csv_folder, file_name)
            df = pd.read_csv(file_path)

            # Calculate metrics based on 'target' and 'prediction_label' columns
            true_labels = df['target']
            predicted_labels = df['prediction_label']

            # Update the categories
            categories = sorted(list(set(true_labels)))
            num_categories = len(categories)

            # Create a confusion matrix for the categories
            confusion_mat = confusion_matrix(true_labels, predicted_labels, labels=categories)

            # Write file name and confusion matrix to the text file
            txt_file.write("File: {}\n".format(file_name))
            txt_file.write("Confusion Matrix:\n")
            txt_file.write(tabulate(confusion_mat, headers=categories, showindex=categories, tablefmt='grid'))
            txt_file.write("\n")

            # Initialize a list to hold metrics for each category
            metrics = []

            # Calculate metrics for each category
            for i in range(num_categories):
                category = categories[i]
                tp = confusion_mat[i, i]
                tn = sum([confusion_mat[j, k] for j in range(num_categories) if j != i for k in range(num_categories) if k != i])
                fp = sum([confusion_mat[j, i] for j in range(num_categories) if j != i])
                fn = sum([confusion_mat[i, j] for j in range(num_categories) if j != i])

                accuracy = (tp + tn) / (tp + tn + fp + fn)
                precision = tp / (tp + fp)
                f1 = 2 * tp / (2 * tp + fp + fn)

                metrics.append({
                    'Category': category,
                    'TP': tp,
                    'TN': tn,
                    'FP': fp,
                    'FN': fn,
                    'Accuracy': accuracy,
                    'Precision': precision,
                    'F1 Score': f1
                })

            # Write category metrics to the text file
            txt_file.write("Category Metrics:\n")
            txt_file.write(tabulate(metrics, headers='keys', tablefmt='grid'))
            txt_file.write("\n")

            # Calculate overall metrics and write them in tabular format
            overall_metrics = calculate_overall_metrics(confusion_mat, num_categories)
            txt_file.write("Overall Metrics:\n")
            txt_file.write(tabulate([overall_metrics], headers='keys', tablefmt='grid'))
            txt_file.write("\n")

def calculate_overall_metrics(confusion_mat, num_categories):
    overall_tp = sum([confusion_mat[i, i] for i in range(num_categories)])
    overall_tn = sum([confusion_mat[i, j] for i in range(num_categories) for j in range(num_categories) if i != j])
    overall_fp = sum([confusion_mat[j, i] for i in range(num_categories) for j in range(num_categories) if i != j])
    overall_fn = sum([confusion_mat[i, j] for i in range(num_categories) for j in range(num_categories) if i != j])

    overall_accuracy = (overall_tp + overall_tn) / (overall_tp + overall_tn + overall_fp + overall_fn)
    overall_precision = overall_tp / (overall_tp + overall_fp)
    overall_f1 = 2 * overall_tp / (2 * overall_tp + overall_fp + overall_fn)

    return {
        'Overall TP': overall_tp,
        'Overall TN': overall_tn,
        'Overall FP': overall_fp,
        'Overall FN': overall_fn,
        'Overall Accuracy': overall_accuracy,
        'Overall Precision': overall_precision,
        'Overall F1 Score': overall_f1
    }

def visualize_data(range_1, range_2):
    input_file = input("File name:")

    # Step 1: Load CSV file
    extension='.csv'
    folder_extension='graphing_data'
    # Get the absolute path of the current script
    script_path = os.path.abspath(__file__)
    # Get the directory containing the script
    script_directory = os.path.dirname(script_path)
    # Combine the script directory with the output file name
    complete_file_path = os.path.join(script_directory, folder_extension ,input_file+extension)

    df = pd.read_csv(complete_file_path)[range_1:range_2]

    # Step 2: Create scatter plots
    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=False)

    # Upper scatter plot (ampl_orig colored by target_3k)
    target_colors = {0: 'g', 1: 'r'}

    for target in df['target'].unique():
        mask = df['target'] == target
        ax1.scatter(df.index[mask], df.loc[mask, 'ampl_orig'], color=target_colors[target], label=f'target={target}', s=5)
        ax1.plot(df['ampl_orig'], linewidth=0.25)

    ax1.set_ylabel('Amplitude [dB]')
    ax1.legend()

    # Lower scatter plot (ampl_orig colored by prediction_label)
    prediction_colors = {0: 'g', 1: 'r'}

    for prediction in df['prediction_label'].unique():
        mask = df['prediction_label'] == prediction
        ax2.scatter(df.index[mask], df.loc[mask, 'ampl_orig'], color=prediction_colors[prediction], label=f'prediction_label={prediction}', s=5)
        ax2.plot(df['ampl_orig'], linewidth=0.25)

    ax2.set_xlabel('Index [minute]')
    ax2.set_ylabel('Amplitude [dB]')
    ax2.legend()

    # Display the plot
    plt.tight_layout()
    plt.show()

# ---------------- Execution of Classification and Post-processing Functions ---------------- #

# If the ML modelling was already done once, and just the visualization is needed
# then the functions can be commented-out except for the visualize_data function

# Classification function
classification()

# Transform and save CSV
decode_trans_rec_data('output_predictions.csv')

# Create the separate CSV files
unmerge_trans_rec_pairs()

# Calculate metrics for each file and save results to the output text file
calculate_evaluation_metrics()

# Call the graphing function (modify the range as needed)
visualize_data(range_1, range_2)
